/*
Program: cipher-gen
File: cipher-gen.cpp
Author: Martin Voros
*/

#include <iostream>
#include <fstream>
#include <vector>
#include <string>
#include <getopt.h>

/*SYNAPS headers*/
#include "synaps/mpol.h"
#include "synaps/arithm/QQ.h"
#include "synaps/linalg/VectDse.h"
#include "synaps/linalg.h"
#include "synaps/arithm/Zp.h"

#include "cipher-gen.h"

using namespace std;

/*Global vars*/
char * output=NULL;
char * input=NULL;
bool logic;

char * func;
vector<matrix_t> lfsr;  // all ciphers LFSRs
vector<lkey_t> key;
vector<interval> range; 
int range_max;
mpol_t eq("0"); // equation
coeff_t res;
int nlfsr = 0;
int nstate = 0;
int bit_size = 0; // total LFSRs size
vector<int> lfsr_size; // size of each LFSR
vector<matrix_t> trans; // LFSR transitions
vector<lkey_t> trans2; // LFSR transitions for nums :-(
vector<unit> combinator;
bool verbose = false;

int main (int argc, char**argv) 
{
	/*Parsing input parameters*/
	arguments(argc, argv);
	load_lfsr(input);
	init_combinator(func);
	init_cipher(); 
	
	
	/*Generating equations*/
	ofstream out (output);
	if (!out.is_open()) {
		cerr << "ERROR: Unable to open output file: " << output << endl;
		exit(1);
	} 
	for (int i=0; i<=range_max; i++) {
		if (verbose) print_state();
		
		if (inter_look_up(i)) {
			eval();
			if (logic) {
				print_logic(out, eq);
				out << "=" << res << '.' << endl;
			} else
				out << eq << "=" << res << endl;
		} 
		make_step();
	}
	cout << "DONE!" << endl;
	return 0;
}

/*
 * Init LFSRs
 */
void 
init_cipher() 
{
	int counter=0;

	if (verbose) cout << " Cipher initialization.... ";
	lfsr.resize(nlfsr); 
	for (int i=0; i < nlfsr; i++) {
		matrix_t tmp(1,lfsr_size[i]);
		for (int j=0; j<tmp.nbcol(); j++) {
			string s = "x" + itos(counter);
			mpol_t p(s.data());
			tmp(0,j)=p;
			counter++;
		}
		lfsr.at(i)=tmp;
	}
	if (verbose) cout << " DONE" << endl;
}

void
init_combinator(char * func)
{
	mpol_t a(func);
	unit foo;
	for(mpol_t::iterator it =a.begin(); it !=a.end();it++) {
		foo.index.clear();
		foo.exp.clear();			
		foo.constant = it->coeff();
		for ( int i=0; i<=it->nvars(); i++) {
			if ( it->rep()[i] > 0 && i > bit_size ) {
				cerr << "ERROR: index x" << i << " in combinator out of range." << endl;
				exit(1);
			}			
			if ( it->rep()[i] > 0) {
				foo.index.push_back(i);
				foo.exp.push_back(it->rep()[i]);
			}
		}
		combinator.push_back(foo);
	}
 	if (verbose) print_combinator();
}

void
make_step()
{
	for (int i=0; i<nlfsr; i++) {
		lfsr[i] = lfsr[i] * trans.at(i);
		key[i] = key[i] * trans2.at(i);
	}
	nstate++;
}

void 
eval()
{
	eq = 0;
	res = 0;
	for (int i=0; i < combinator.size(); i++) {
		mpol_t tmp = combinator.at(i).constant;
		coeff_t tr = combinator.at(i).constant;
		for ( int j=0; j < combinator.at(i).index.size(); j++) {
			int x=0;
			int y = combinator.at(i).index.at(j);
			while ( y >= lfsr_size.at(x) ) {
				y -= lfsr_size.at(x);
				x++;				
			}
			if ( combinator.at(i).exp.at(j) > 1 ) {
				int m = combinator.at(i).exp.at(j);
				mpol_t r("1");
				coeff_t r2=1;
				mpol_t e=lfsr.at(x)(0,y);
				coeff_t e2=key.at(x)(0,y);
				while ( m!= 0) {
					if (m%2) {
						m--; m /= 2; r*=e; r2*=e2; e=e*e; e2=e2*e2;
					} else {
						m /= 2;	e = e*e; e2 = e2*e2;
					}
				}
				tmp *= r;
				tr *= r2;
			} else {
				tmp *= lfsr.at(x)(0,y);
				tr *= key.at(x)(0,y);
			}
		}
		eq += tmp;
		res += tr;
	}
	
}

void 
print_state()
{
	cout << " Cipher state ["<< nstate <<"]: " << endl;
	for (int i=0; i < lfsr.size(); i++) {
		cout << "LFSR" << i << ": " << lfsr[i];
		cout << "BITS" << i << ": " << key[i];
	}
}

void 
print_trans()
{
	cout << " Transition" << endl;
	for (int i=0; i < trans.size(); i++) {
		cout << "LFSR" << i << ": " << endl << trans.at(i) << endl;
	}
}

void
print_combinator()
{
	cout << " Combinator" << endl << " ";
	for (int i = 0; i<combinator.size(); i++) {
		bool first = false;
		if ( combinator.at(i).constant!=1 )
			cout << combinator.at(i).constant;
		else first=true;
		for (int j=0; j<combinator.at(i).index.size(); j++) {
			if (first) {
				cout << "x";
				first=false;
			} else cout << "*x";
			cout << combinator.at(i).index.at(j);
			if (combinator.at(i).exp.at(j) != 1)
				cout <<"^" << combinator.at(i).exp.at(j);
		}
		if ( (i+1) != combinator.size() ) cout << " + ";
	}
	cout << endl;
}

void 
print_key() 
{
	cout << " Key " << endl;
	for (int i=0; i < key.size(); i++) {
		cout << "LFSR" << i << key.at(i);
	}
}

/*
 * Prints generated equation in logic format  /\ & xor
 */

void 
print_logic(ofstream &out, mpol_t &i)
{
	mpol_t::iterator it, pt;

	for(it=i.begin(); it !=i.end(); it++) {
		for (int j=0; j<=it->nvars(); j++)
			if (j+1 > it->nvars()) { 
				if (it->rep()[j] > 0) 
					out << j+1;
			} else {
				if (it->rep()[j] > 0) 
					out << j+1 << "*";
			} 
		pt=it; pt++;
		if (pt != i.end())
			out << "+";
	}
}

string 
itos(int i)	
{
	stringstream s;
	s << i;
	return s.str();
}

void usage ()
{
	cout <<"cipher-gen (" << MYVERSION << ")" << endl;
	cout <<"Usage: cipher-gen [options]"<< endl
	<<"	-h : this help message" << endl
	<<"	-V : print version info" << endl
	<<"	-v : verbose mode " << endl
	<<"	-r : range '1-5,8,9-15' [default 1]" << endl
	<<"	-o : output file [required]" << endl
	<<"	-c : combinator 'x1*x2+3*x6+4' [required]" << endl
	<<"	-i : LFSR config file [required]" << endl
	<<"	-l : logic output format [default false] - works only for GF(2)" << endl
       <<endl;
}

void
arguments(int argc, char* argv[])
{
	int next;
	bool r = false;
	const char* const short_options = "vVho:c:li:r:";
	do {
	next = getopt(argc, argv, short_options);
	switch (next) {
		case '?' : usage(); break;
		case 'V' : cout <<"cipher-gen (" << MYVERSION << ")" << endl;
			   exit(1);
			   break;
		case 'r' : set_range(optarg);
			   r=true;
			   break;
		case 'o' : output=optarg; break;
		case 'v' : verbose = true; break;
		case 'i' : input=optarg; break;
		case 'l' : logic=true; break;
		case 'c' : func=optarg; break;
		case 'h' : usage(); 
			   exit(1);
			   break;
	}
	} while (next != -1 );
	/*Value chceck*/
	if (output == NULL || func == NULL || input == NULL) {
		cerr << "ERROR: Required parameter is missing." << endl;
		usage();
		exit(1);
	} 
	if (!r) {
		interval intr = {1,1};
		range_max=1;
		range.push_back(intr);
	}
}

void
load_lfsr(char * input)
{
	fstream in(input, ios::in);
	if (in.is_open()){
		int foo=0;
		char line[LINE] = "";
		while (!in.eof()) {
			// getting LFSR size
			int size = 0;
			in.getline(line,LINE);
			istringstream keybits(line);
			keybits >> size;
			if (size > 0) {
				// setting key
				lkey_t k(1,size);
				for(int i=0; i< size && keybits.good(); i++) {
					foo=0;
					keybits >> foo;
					k(0,i)=foo;	
				}
				key.push_back(k);
				bit_size += size;	
				lfsr_size.push_back(size);
				// getting lfsr
				in.getline(line,LINE);
				matrix_t M(size,size);
				lkey_t M2(size,size);
				istringstream matrix(line);
				for(int i=0; i< size && matrix.good(); i++)
    					for(int j=0; j< size && matrix.good(); j++) {
						foo=0;
      						matrix >> foo;	
						mpol_t p(foo);			
						M(i,j)=p;
						M2(i,j)=foo;
				}	
				trans.push_back(M);
				trans2.push_back(M2);
				nlfsr++;
			}
		}
	}
  	else {
		cerr << "ERROR: Unable to open input file" << input << endl;
  		exit(1);
	}
	in.close();
	if (verbose) {
		print_trans();
		print_key();
	}
}

bool
inter_look_up(int i) {
	if (range.size() > 0) {
		int j=0;
		while ( i >= range.at(j).a && j<range.size() ) {
			if ( i <= range.at(j).b ) return true;
			else j++;
		}
		return false;	
	} else {
		return false;	
	}
}

void
set_range(char * s)
{
	int a,b;
	char c;
	interval tmp;

	istringstream arg(s);	
	while ( !arg.eof() ) {
		arg >> a;	
		c=arg.peek();
		switch (c) {
			case ',': 
				arg.ignore(1,EOF); 
				tmp.a=a; tmp.b=a;
				range.push_back(tmp);
				range_max=a;
				break;
			case '-':
				arg.ignore(1,EOF); 
				arg >> b;
				if (b >= a) {
					tmp.a=a;
					tmp.b=b;
					range.push_back(tmp);	
					range_max=b;
				}
				break;
			default :
				cerr << "ERROR: Invalid eq range" << endl;
				exit(1);
		}
	}	
}
